//
//  Simple class hierarchy to manage various "locking" schemes
//
//      Interlock           - Use atomic intstructions to manage an integer state
//      NoLock              - Don't do any locking, useful for comparison testing
//      CriticalSection     - Use MS critical section
//
//  Project: Thread optimized Ring Buffer 
//  Author: Dennis Lang
//  lang.dennis @ comcast.net 
//  http://home.comcast.net/~lang.dennis/

#include <windows.h>
#pragma once

// ---------------------------------------------------------------------------
class LockBase
{
public:
    LockBase() {}
    virtual void Lock() = 0;
    virtual void Release() = 0;
};


// ---------------------------------------------------------------------------
class LockIt
{
public:
    LockIt(LockBase& lock): pLock(&lock)
    {
        pLock->Lock();
    }
    ~LockIt()
    {
        pLock->Release();
    }
private:
    LockBase* pLock;
};


// ---------------------------------------------------------------------------
class Interlock : public LockBase
{
public:
    Interlock() : m_value(0) {}

    typedef LONG Atomic_t;

    bool Exchange(Atomic_t v)
    {
        Atomic_t compare = !v;
        return (Atomic_t)InterlockedCompareExchangeAcquire(&m_value, v, compare) == (Atomic_t)true;
    }

    void Lock()
    {
        while (Exchange(true))
        {}
    }

    void Release()
    {
        m_value = false;
    }

private:
    volatile Atomic_t m_value;
};

// ---------------------------------------------------------------------------
class NoLock : public LockBase
{
public:
    NoLock() {}

    void Lock()
    {
    }

    void Release()
    {
    }
};


// ---------------------------------------------------------------------------
class CriticalSection : public LockBase
{
public:
    CriticalSection()  { InitializeCriticalSection(&m_value); }
    ~CriticalSection() { DeleteCriticalSection(&m_value); }

    void Lock()        { EnterCriticalSection(&m_value);  }
    void Release()     { LeaveCriticalSection(&m_value); }

private:
    CRITICAL_SECTION m_value;
};

