// -----------------------------------------------------------------------------
//
// The following code is not required to run the test assuming you comment out the 
// calls to MemoryHook in TestBufferSpeed.cpp.
//
// This code is an attempt to monitor all memory allocations while a buffer is being
// tested so it can be reported along with its performance.
//
// NOTE: The memory hook only works on 32bit builds.  Under x64 builds (_M_X64)
// I overload "new" and "delete" to track memory usage.
//
// Hook HeapAlloc/Realloc/Free to track memory allocations
//
// Following code was borrowed from this link:
// http://src.chromium.org/svn/trunk/src/tools/memory_watcher/memory_hook.cc
//
// -----------------------------------------------------------------------------

#include <windows.h>

// -----------------------------------------------------------------------------
class MemoryHook 
{
public:
    // Returns true is memory allocations and deallocations
    // are being traced.
    static bool IsHooked() { return m_sIsHooked != NULL; }

    static long GetByteCnt() { return m_sByteCnt; }
    static long GetMaxByteCnt() { return m_sMaxByteCnt; }

    static void ResetByteCnt() 
    { m_sByteCnt = m_sMaxByteCnt = 0; }

#ifdef _M_X64
    static bool Hook() { return false; }
    static bool Unhook() { return false; }
#else

    // Enable memory tracing. 
    static bool Hook();

    // Disables memory tracing.
    static bool Unhook();

#pragma pack(push, MEMORY_ALLOCATION_ALIGNMENT)

    static const int sAllocCnt = 10;
    static const int sAllocChunkSz = 32;
    static char m_sMallocBufs[sAllocCnt][sAllocChunkSz];
    static unsigned int m_sAllocIdx;

#pragma pack(pop, MEMORY_ALLOCATION_ALIGNMENT)
    static void* Alloc(size_t size) 
    {
        // return malloc(size);
        assert(m_sAllocIdx < sAllocCnt);
        assert(size <= sAllocChunkSz);
        return m_sMallocBufs[m_sAllocIdx++];
    }	

    static void Free(void* ptr) 
    {
        // free(ptr);
    }
#endif

    static bool m_sIsHooked;
    static long m_sByteCnt;
    static long m_sMaxByteCnt;
};


// -----------------------------------------------------------------------------
bool MemoryHook::m_sIsHooked = false;
long MemoryHook::m_sByteCnt = 0;
long MemoryHook::m_sMaxByteCnt = 0;

#ifdef _M_X64

// Global new
void *operator new(size_t size)
{
    void* ptr  =  malloc(size);
    MemoryHook::m_sByteCnt += size;
    if (MemoryHook::m_sByteCnt > MemoryHook::m_sMaxByteCnt)
        MemoryHook::m_sMaxByteCnt = MemoryHook::m_sByteCnt;
    return ptr;
}
   
// Global delete
void operator delete(void* ptr)
{
    if (ptr != NULL)
    {
        MemoryHook::m_sByteCnt = _msize(ptr);
        free(ptr);
    }
}

#else

char MemoryHook::m_sMallocBufs[sAllocCnt][32];
unsigned int MemoryHook::m_sAllocIdx = 0;


// -----------------------------------------------------------------------------
// Calls GetProcAddress, but casts to the correct type.
#define GET_PROC_ADDRESS(hmodule, name) \
    ( (Type_##name)(::GetProcAddress(hmodule, #name)) )

// Macro to declare Patch functions.
#define DECLARE_PATCH(name) Patch<Type_##name> patch_##name

// Macro to install Patch functions.
#define INSTALL_PATCH(name)  do {                                       \
    patch_##name.set_original(GET_PROC_ADDRESS(hkernel32, ##name));       \
    patch_##name.Install(&Perftools_##name);                              \
} while (0)

// Macro to install Patch functions.
#define INSTALL_NTDLLPATCH(name)  do {                                  \
    patch_##name.set_original(GET_PROC_ADDRESS(hntdll, ##name));          \
    patch_##name.Install(&Perftools_##name);                              \
} while (0)

// Macro to uninstall Patch functions.
#define UNINSTALL_PATCH(name) patch_##name.Uninstall();



#include "mini_disassembler.h"

// -----------------------------------------------------------------------------
// Windows APIs to be hooked

// HeapAlloc routines
typedef HANDLE (WINAPI *Type_HeapCreate)(
    DWORD flOptions, SIZE_T dwInitialSize, SIZE_T dwMaximumSize);
typedef BOOL (WINAPI *Type_HeapDestroy)(HANDLE hHeap);
typedef LPVOID (WINAPI *Type_HeapAlloc)(
    HANDLE hHeap, DWORD dwFlags, DWORD_PTR dwBytes);
typedef LPVOID (WINAPI *Type_HeapReAlloc)(
    HANDLE hHeap, DWORD dwFlags, LPVOID lpMem, SIZE_T dwBytes);
typedef BOOL (WINAPI *Type_HeapFree)(
    HANDLE hHeap, DWORD dwFlags, LPVOID lpMem);





// -----------------------------------------------------------------------------
// Possible results of patching/unpatching
enum SideStepError {
    SIDESTEP_SUCCESS = 0,
    SIDESTEP_INVALID_PARAMETER,
    SIDESTEP_INSUFFICIENT_BUFFER,
    SIDESTEP_JUMP_INSTRUCTION,
    SIDESTEP_FUNCTION_TOO_SMALL,
    SIDESTEP_UNSUPPORTED_INSTRUCTION,
    SIDESTEP_NO_SUCH_MODULE,
    SIDESTEP_NO_SUCH_FUNCTION,
    SIDESTEP_ACCESS_DENIED,
    SIDESTEP_UNEXPECTED,
};

#define SIDESTEP_TO_HRESULT(error) \
    MAKE_HRESULT(SEVERITY_ERROR, FACILITY_NULL, error)

// Implements a patching mechanism that overwrites the first few bytes of
// a function preamble with a jump to our hook function, which is then
// able to call the original function via a specially-made preamble-stub
// that imitates the action of the original preamble.
//
// NOTE:  This patching mechanism should currently only be used for
// non-production code, e.g. unit tests, because it is not threadsafe.
// See the TODO in preamble_patcher_with_stub.cc for instructions on what
// we need to do before using it in production code; it's fairly simple
// but unnecessary for now since we only intend to use it in unit tests.
//
// To patch a function, use either of the typesafe Patch() methods.  You
// can unpatch a function using Unpatch().
//
// Typical usage goes something like this:
// @code
// typedef int (*MyTypesafeFuncPtr)(int x);
// MyTypesafeFuncPtr original_func_stub;
// int MyTypesafeFunc(int x) { return x + 1; }
// int HookMyTypesafeFunc(int x) { return 1 + original_func_stub(x); }
//
// void MyPatchInitializingFunction() {
//   original_func_stub = PreamblePatcher::Patch(
//              MyTypesafeFunc, HookMyTypesafeFunc);
//   if (!original_func_stub) {
//     // ... error handling ...
//   }
//
//   // ... continue - you have patched the function successfully ...
// }
// @endcode
//
// Note that there are a number of ways that this method of patching can
// fail.  The most common are:
//    - If there is a jump (jxx) instruction in the first 5 bytes of
//    the function being patched, we cannot patch it because in the
//    current implementation we do not know how to rewrite relative
//    jumps after relocating them to the preamble-stub.  Note that
//    if you really really need to patch a function like this, it
//    would be possible to add this functionality (but at some cost).
//    - If there is a return (ret) instruction in the first 5 bytes
//    we cannot patch the function because it may not be long enough
//    for the jmp instruction we use to inject our patch.
//    - If there is another thread currently executing within the bytes
//    that are copied to the preamble stub, it will crash in an undefined
//    way.
//
// If you get any other error than the above, you're either pointing the
// patcher at an invalid instruction (e.g. into the middle of a multi-
// byte instruction, or not at memory containing executable instructions)
// or, there may be a bug in the disassembler we use to find
// instruction boundaries.
//
// NOTE:  In optimized builds, when you have very trivial functions that
// the compiler can reason do not have side effects, the compiler may
// reuse the result of calling the function with a given parameter, which
// may mean if you patch the function in between your patch will never get
// invoked.  See preamble_patcher_test.cc for an example.
class PreamblePatcher 
{
public:

    // This is a typesafe version of RawPatch(), identical in all other
    // ways than it takes a template parameter indicating the type of the
    // function being patched.
    //
    // @param T The type of the function you are patching. Usually
    // you will establish this type using a typedef, as in the following
    // example:
    // @code
    // typedef BOOL (WINAPI *MessageBoxPtr)(HWND, LPCTSTR, LPCTSTR, UINT);
    // MessageBoxPtr original = NULL;
    // PreamblePatcher::Patch(MessageBox, Hook_MessageBox, &original);
    // @endcode
    template <class T>
    static SideStepError Patch(
        T target_function,
        T replacement_function,
        T* original_function_stub) 
    {
        // NOTE: casting from a function to a pointer is contra the C++
        //       spec.  It's not safe on IA64, but is on i386.  We use
        //       a C-style cast here to emphasize this is not legal C++.
        return RawPatch((void*)(target_function),
            (void*)(replacement_function),
            (void**)(original_function_stub));
    }

    // Patches a named function imported from the named module using
    // preamble patching.  Uses RawPatch() to do the actual patching
    // work.
    //
    // @param T The type of the function you are patching.  Must
    // exactly match the function you specify using module_name and
    // function_name.
    //
    // @param module_name The name of the module from which the function
    // is being imported.  Note that the patch will fail if this module
    // has not already been loaded into the current process.
    //
    // @param function_name The name of the function you wish to patch.
    //
    // @param replacement_function Your replacement function which
    // will be called whenever code tries to call the original function.
    //
    // @param original_function_stub Pointer to memory that should receive a
    // pointer that can be used (e.g. in the replacement function) to call the
    // original function, or NULL to indicate failure.
    //
    // @return One of the EnSideStepError error codes; only SIDESTEP_SUCCESS
    // indicates success.
    template <class T>
    static SideStepError Patch(
        LPCTSTR module_name,
        LPCSTR function_name,
        T replacement_function,
        T* original_function_stub) 
    {
        ASSERT1(module_name && function_name);
        if (!module_name || !function_name) 
        {
            ASSERT(false,
                "You must specify a module name and function name.");
            return SIDESTEP_INVALID_PARAMETER;
        }

        HMODULE module = ::GetModuleHandle(module_name);
        ASSERT1(module != NULL);
        if (!module) 
        {
            ASSERT(false, "Invalid module name.");
            return SIDESTEP_NO_SUCH_MODULE;
        }

        FARPROC existing_function = ::GetProcAddress(module, function_name);
        if (!existing_function) 
        {
            return SIDESTEP_NO_SUCH_FUNCTION;
        }

        // NOTE: casting from a function to a pointer is contra the C++
        //       spec.  It's not safe on IA64, but is on i386.  We use
        //       a C-style cast here to emphasize this is not legal C++.
        return RawPatch((void*)existing_function, (void*)replacement_function,
            (void**)(original_function_stub));
    }

    // Patches a function by overwriting its first few bytes with
    // a jump to a different function.  This is the "worker" function
    // for each of the typesafe Patch() functions.  In most cases,
    // it is preferable to use the Patch() functions rather than
    // this one as they do more checking at compile time.
    //
    // @param target_function A pointer to the function that should be
    // patched.
    //
    // @param replacement_function A pointer to the function that should
    // replace the target function.  The replacement function must have
    // exactly the same calling convention and parameters as the original
    // function.
    //
    // @param original_function_stub Pointer to memory that should receive a
    // pointer that can be used (e.g. in the replacement function) to call the
    // original function, or NULL to indicate failure.
    //
    // @param original_function_stub Pointer to memory that should receive a
    // pointer that can be used (e.g. in the replacement function) to call the
    // original function, or NULL to indicate failure.
    //
    // @return One of the EnSideStepError error codes; only SIDESTEP_SUCCESS
    // indicates success.
    //
    // @note The preamble-stub (the memory pointed to by
    // *original_function_stub) is allocated on the heap, and (in
    // production binaries) never destroyed, resulting in a memory leak.  This
    // will be the case until we implement safe unpatching of a method.
    // However, it is quite difficult to unpatch a method (because other
    // threads in the process may be using it) so we are leaving it for now.
    // See however UnsafeUnpatch, which can be used for binaries where you
    // know only one thread is running, e.g. unit tests.
    static SideStepError RawPatch(
        void* target_function,
        void* replacement_function,
        void** original_function_stub);

    // Unpatches target_function and deletes the stub that previously could be
    // used to call the original version of the function.
    //
    // DELETES the stub that is passed to the function.
    //
    // @param target_function Pointer to the target function which was
    // previously patched, i.e. a pointer which value should match the value
    // of the symbol prior to patching it.
    //
    // @param replacement_function Pointer to the function target_function
    // was patched to.
    //
    // @param original_function_stub Pointer to the stub returned when
    // patching, that could be used to call the original version of the
    // patched function.  This function will also delete the stub, which after
    // unpatching is useless.
    //
    // If your original call was
    //    origptr = Patch(VirtualAlloc, MyVirtualAlloc)
    // then to undo it you would call
    //    Unpatch(VirtualAlloc, MyVirtualAlloc, origptr);
    //
    // @return One of the EnSideStepError error codes; only SIDESTEP_SUCCESS
    // indicates success.
    static SideStepError Unpatch(
        void* target_function,
        void* replacement_function,
        void* original_function_stub);

private:

    // Patches a function by overwriting its first few bytes with
    // a jump to a different function.  This is similar to the RawPatch
    // function except that it uses the stub allocated by the caller
    // instead of allocating it.
    //
    // We call VirtualProtect to make the
    // target function writable at least for the duration of the call.
    //
    // @param target_function A pointer to the function that should be
    // patched.
    //
    // @param replacement_function A pointer to the function that should
    // replace the target function.  The replacement function must have
    // exactly the same calling convention and parameters as the original
    // function.
    //
    // @param preamble_stub A pointer to a buffer where the preamble stub
    // should be copied. The size of the buffer should be sufficient to
    // hold the preamble bytes.
    //
    // @param stub_size Size in bytes of the buffer allocated for the
    // preamble_stub
    //
    // @param bytes_needed Pointer to a variable that receives the minimum
    // number of bytes required for the stub.  Can be set to NULL if you're
    // not interested.
    //
    // @return An error code indicating the result of patching.
    static SideStepError RawPatchWithStubAndProtections(
        void* target_function,
        void *replacement_function,
        unsigned char* preamble_stub,
        unsigned long stub_size,
        unsigned long* bytes_needed);

    // A helper function used by RawPatchWithStubAndProtections -- it does
    // everything but the VirtualProtect wsork.  Defined in
    // preamble_patcher_with_stub.cc.
    static SideStepError RawPatchWithStub(
        void* target_function,
        void *replacement_function,
        unsigned char* preamble_stub,
        unsigned long stub_size,
        unsigned long* bytes_needed);
};


// Maximum size of the preamble stub. We overwrite at least the first 5
// bytes of the function. Considering the worst case scenario, we need 4
// bytes + the max instruction size + 5 more bytes for our jump back to
// the original code. With that in mind, 32 is a good number :)
const unsigned int MAX_PREAMBLE_STUB_SIZE = 32;

// Definitions of assembly statements we need
#define ASM_JMP32REL 0xE9
#define ASM_INT3 0xCC

SideStepError PreamblePatcher::RawPatchWithStub(
    void* target_function,
    void *replacement_function,
    unsigned char* preamble_stub,
    unsigned long stub_size,
    unsigned long* bytes_needed) 
{
    if ((NULL == target_function) ||
        (NULL == replacement_function) ||
        (NULL == preamble_stub)) 
    {
        // ASSERT(false, "Invalid parameters - either pTargetFunction or "
        //              "pReplacementFunction or pPreambleStub were NULL.");
        return SIDESTEP_INVALID_PARAMETER;
    }

    // TODO(V7:joi) Siggi and I just had a discussion and decided that both
    // patching and unpatching are actually unsafe.  We also discussed a
    // method of making it safe, which is to freeze all other threads in the
    // process, check their thread context to see if their eip is currently
    // inside the block of instructions we need to copy to the stub, and if so
    // wait a bit and try again, then unfreeze all threads once we've patched.
    // Not implementing this for now since we're only using SideStep for unit
    // testing, but if we ever use it for production code this is what we
    // should do.
    //
    // NOTE: Stoyan suggests we can write 8 or even 10 bytes atomically using
    // FPU instructions, and on newer processors we could use cmpxchg8b or
    // cmpxchg16b. So it might be possible to do the patching/unpatching
    // atomically and avoid having to freeze other threads.  Note though, that
    // doing it atomically does not help if one of the other threads happens
    // to have its eip in the middle of the bytes you change while you change
    // them.
    unsigned char* target = reinterpret_cast<unsigned char*>(target_function);

    // First, deal with a special case that we see with functions that
    // point into an IAT table (including functions linked statically
    // into the application): these function already starts with
    // ASM_JMP32REL.  For instance, malloc() might be implemented as a
    // JMP to __malloc().  In that case, we replace the destination of
    // the JMP (__malloc), rather than the JMP itself (malloc).  This
    // way we get the correct behavior no matter how malloc gets called.
    if (target[0] == ASM_JMP32REL)
    {
        // target[1-4] holds the place the jmp goes to, but it's
        // relative to the next instruction.
        int relative_offset;   // Windows guarantees int is 4 bytes
        // ASSERT1(sizeof(relative_offset) == 4);
        memcpy(reinterpret_cast<void*>(&relative_offset),
            reinterpret_cast<void*>(target + 1), 4);
        // I'd like to just say "target = target + 5 + relative_offset" here, but
        // I can't, because the new target will need to have its protections set.
        return RawPatchWithStubAndProtections(target + 5 + relative_offset,
            replacement_function, preamble_stub,
            stub_size, bytes_needed);
    }

    // Let's disassemble the preamble of the target function to see if we can
    // patch, and to see how much of the preamble we need to take.  We need 5
    // bytes for our jmp instruction, so let's find the minimum number of
    // instructions to get 5 bytes.
    MiniDisassembler disassembler;
    unsigned int preamble_bytes = 0;
    while (preamble_bytes < 5) 
    {
        InstructionType instruction_type =
            disassembler.Disassemble(target + preamble_bytes, preamble_bytes);

        if (IT_JUMP == instruction_type) 
        {
            // ASSERT(false, "Unable to patch because there is a jump instruction in the first 5 bytes.");
            return SIDESTEP_JUMP_INSTRUCTION;
        } 
        else if (IT_RETURN == instruction_type) 
        {
            // ASSERT(false, "Unable to patch because function is too short");
            return SIDESTEP_FUNCTION_TOO_SMALL;
        } 
        else if (IT_GENERIC != instruction_type) 
        {
            // ASSERT(false, "Disassembler encountered unsupported instruction (either unused or unknown)");
            return SIDESTEP_UNSUPPORTED_INSTRUCTION;
        }
    }

    if (NULL != bytes_needed)
        *bytes_needed = preamble_bytes + 5;

    // Inv: cbPreamble is the number of bytes (at least 5) that we need to take
    // from the preamble to have whole instructions that are 5 bytes or more
    // in size total. The size of the stub required is cbPreamble + size of
    // jmp (5)
    if (preamble_bytes + 5 > stub_size) 
    {
        assert(false);
        return SIDESTEP_INSUFFICIENT_BUFFER;
    }

    // First, copy the preamble that we will overwrite.
    memcpy(reinterpret_cast<void*>(preamble_stub),
        reinterpret_cast<void*>(target), preamble_bytes);

    // Now, make a jmp instruction to the rest of the target function (minus the
    // preamble bytes we moved into the stub) and copy it into our preamble-stub.
    // find address to jump to, relative to next address after jmp instruction
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4244)
#endif
    int relative_offset_to_target_rest
        = ((reinterpret_cast<unsigned char*>(target) + preamble_bytes) -
        (preamble_stub + preamble_bytes + 5));
#ifdef _MSC_VER
#pragma warning(pop)
#endif
    // jmp (Jump near, relative, displacement relative to next instruction)
    preamble_stub[preamble_bytes] = ASM_JMP32REL;
    // copy the address
    memcpy(reinterpret_cast<void*>(preamble_stub + preamble_bytes + 1),
        reinterpret_cast<void*>(&relative_offset_to_target_rest), 4);

    // Inv: preamble_stub points to assembly code that will execute the
    // original function by first executing the first cbPreamble bytes of the
    // preamble, then jumping to the rest of the function.

    // Overwrite the first 5 bytes of the target function with a jump to our
    // replacement function.
    // (Jump near, relative, displacement relative to next instruction)
    target[0] = ASM_JMP32REL;

    // Find offset from instruction after jmp, to the replacement function.
#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable:4244)
#endif
    int offset_to_replacement_function =
        reinterpret_cast<unsigned char*>(replacement_function) -
        reinterpret_cast<unsigned char*>(target) - 5;
#ifdef _MSC_VER
#pragma warning(pop)
#endif
    // complete the jmp instruction
    memcpy(reinterpret_cast<void*>(target + 1),
        reinterpret_cast<void*>(&offset_to_replacement_function), 4);
    // Set any remaining bytes that were moved to the preamble-stub to INT3 so
    // as not to cause confusion (otherwise you might see some strange
    // instructions if you look at the disassembly, or even invalid
    // instructions). Also, by doing this, we will break into the debugger if
    // some code calls into this portion of the code.  If this happens, it
    // means that this function cannot be patched using this patcher without
    // further thought.
    if (preamble_bytes > 5)
    {
        memset(reinterpret_cast<void*>(target + 5), ASM_INT3, preamble_bytes - 5);
    }

    // Inv: The memory pointed to by target_function now points to a relative
    // jump instruction that jumps over to the preamble_stub.  The preamble
    // stub contains the first stub_size bytes of the original target
    // function's preamble code, followed by a relative jump back to the next
    // instruction after the first cbPreamble bytes.

    return SIDESTEP_SUCCESS;
}

// -----------------------------------------------------------------------------
SideStepError PreamblePatcher::RawPatchWithStubAndProtections(
    void* target_function, 
    void *replacement_function,
    unsigned char* preamble_stub,
    unsigned long stub_size,
    unsigned long* bytes_needed) 
{
    // We need to be able to write to a process-local copy of the first
    // MAX_PREAMBLE_STUB_SIZE bytes of target_function. We may be giving execute
    // privilege to something that doesn't have it, but that's the price to pay
    // for tools.
    DWORD old_target_function_protect = 0;
    BOOL succeeded = ::VirtualProtect(reinterpret_cast<void*>(target_function),
        MAX_PREAMBLE_STUB_SIZE,
        PAGE_EXECUTE_READWRITE,
        &old_target_function_protect);

    if (!succeeded) 
    {
        // ASSERT(false, "Failed to make page containing target function "  "copy-on-write.");
        return SIDESTEP_ACCESS_DENIED;
    }

    SideStepError error_code = RawPatchWithStub(target_function,
        replacement_function,
        preamble_stub,
        stub_size,
        bytes_needed);
    if (SIDESTEP_SUCCESS != error_code) 
    {
        assert(false);
        return error_code;
    }

    // Restore the protection of the first MAX_PREAMBLE_STUB_SIZE bytes of
    // pTargetFunction to what they were before we started goofing around.
    succeeded = ::VirtualProtect(reinterpret_cast<void*>(target_function),
        MAX_PREAMBLE_STUB_SIZE,
        old_target_function_protect,
        &old_target_function_protect);
    if (!succeeded) 
    {
        // ASSERT(false, "Failed to restore protection to target function.");

        // We must not return an error here because the function has actually
        // been patched, and returning an error would likely cause our client
        // code not to unpatch it.  So we just keep going.
    }

    // Flush the instruction cache to make sure the processor doesn't execute the
    // old version of the instructions (before our patch).
    //
    // FlushInstructionCache is actually a no-op at least on single-processor
    // XP machines.  I'm not sure why this is so, but it is, yet I want to keep
    // the call to the API here for correctness in case there is a difference in
    // some variants of Windows/hardware.
    succeeded = ::FlushInstructionCache(::GetCurrentProcess(),
        target_function,
        MAX_PREAMBLE_STUB_SIZE);

    if (!succeeded) 
    {
        // ASSERT(false, "Failed to flush instruction cache.");

        // We must not return an error here because the function has actually
        // been patched, and returning an error would likely cause our client
        // code not to unpatch it.  So we just keep going.
    }

    return SIDESTEP_SUCCESS;
}

// -----------------------------------------------------------------------------
SideStepError PreamblePatcher::RawPatch(
    void* target_function,
    void* replacement_function,
    void** original_function_stub) 
{
    if (!target_function || !replacement_function || !original_function_stub ||
        (*original_function_stub) || target_function == replacement_function) 
    {
        // ASSERT(false, "Preconditions not met");
        return SIDESTEP_INVALID_PARAMETER;
    }

    // @see MAX_PREAMBLE_STUB_SIZE for an explanation of how we arrives at
    // this size
    unsigned char* preamble_stub =
        reinterpret_cast<unsigned char*>(
        MemoryHook::Alloc(sizeof(unsigned char) * MAX_PREAMBLE_STUB_SIZE));

    if (!preamble_stub)
    {
        // ASSERT(false, "Unable to allocate preamble-stub.");
        return SIDESTEP_INSUFFICIENT_BUFFER;
    }

    // Change the protection of the newly allocated preamble stub to
    // PAGE_EXECUTE_READWRITE. This is required to work with DEP (Data
    // Execution Prevention) which will cause an exception if code is executed
    // from a page on which you do not have read access.
    DWORD old_stub_protect = 0;
    BOOL succeeded = VirtualProtect(preamble_stub, MAX_PREAMBLE_STUB_SIZE,
        PAGE_EXECUTE_READWRITE, &old_stub_protect);

    if (!succeeded) 
    {
        // ASSERT(false, "Failed to make page preamble stub read-write-execute.");
        delete[] preamble_stub;
        return SIDESTEP_ACCESS_DENIED;
    }

    SideStepError error_code = RawPatchWithStubAndProtections(target_function,
        replacement_function,
        preamble_stub,
        MAX_PREAMBLE_STUB_SIZE,
        NULL);

    if (SIDESTEP_SUCCESS != error_code) 
    {
        // ASSERT1(false);
        delete[] preamble_stub;
        return error_code;
    }

    *original_function_stub = reinterpret_cast<void*>(preamble_stub);

    // NOTE: For hooking malloc/free, we don't want to use streams which
    // allocate.  Basically, we've hooked malloc, but not necessarily
    // hooked free yet.  To do anything which uses the heap could crash
    // with a mismatched malloc/free!
    //VLOG(1) << "PreamblePatcher::RawPatch successfully patched 0x"
    //        << target_function;

    return SIDESTEP_SUCCESS;
}

// -----------------------------------------------------------------------------
SideStepError PreamblePatcher::Unpatch(
    void* target_function,
    void* replacement_function,
    void* original_function_stub) 
{
    // ASSERT1(target_function && original_function_stub);
    if (!target_function || !original_function_stub) 
    {
        return SIDESTEP_INVALID_PARAMETER;
    }

    // We disassemble the preamble of the _stub_ to see how many bytes we
    // originally copied to the stub.
    MiniDisassembler disassembler;
    unsigned int preamble_bytes = 0;
    while (preamble_bytes < 5) 
    {
        InstructionType instruction_type = disassembler.Disassemble(
            reinterpret_cast<unsigned char*>(original_function_stub) +
            preamble_bytes, preamble_bytes);
        if (IT_GENERIC != instruction_type) 
        {
            ASSERT(false, "Should only have generic instructions in stub!!");
            return SIDESTEP_UNSUPPORTED_INSTRUCTION;
        }
    }

    // Before unpatching, target_function should be a JMP to
    // replacement_function.  If it's not, then either it's an error, or
    // we're falling into the case where the original instruction was a
    // JMP, and we patched the jumped_to address rather than the JMP
    // itself.  (For instance, if malloc() is just a JMP to __malloc(),
    // we patched __malloc() and not malloc().)
    unsigned char* target = reinterpret_cast<unsigned char*>(target_function);
    while (1) 
    {    // we stop when target is a JMP to replacement_function
        if (target[0] != ASM_JMP32REL) 
        {
            ASSERT(false, "target_function does not look like it was patched.");
            return SIDESTEP_INVALID_PARAMETER;
        }
        int relative_offset;   // Windows guarantees int is 4 bytes
        ASSERT1(sizeof(relative_offset) == 4);
        memcpy(reinterpret_cast<void*>(&relative_offset),
            reinterpret_cast<void*>(target + 1), 4);
        unsigned char* jump_to = target + 5 + relative_offset;
        if (jump_to == replacement_function)
            break;
        target = jump_to;      // follow the jmp
    }

    // We need to be able to write to a process-local copy of the first
    // MAX_PREAMBLE_STUB_SIZE bytes of target_function. We may be giving execute
    // privilege to something that doesn't have it, but that's the price to pay
    // for tools.
    DWORD old_target_function_protect = 0;
    BOOL succeeded = ::VirtualProtect(reinterpret_cast<void*>(target),
        MAX_PREAMBLE_STUB_SIZE,
        PAGE_EXECUTE_READWRITE,
        &old_target_function_protect);

    if (!succeeded) 
    {
        ASSERT(false, "Failed to make page containing target function "
            "copy-on-write.");
        return SIDESTEP_ACCESS_DENIED;
    }

    // Replace the first few bytes of the original function with the bytes we
    // previously moved to the preamble stub.
    memcpy(reinterpret_cast<void*>(target),
        original_function_stub, preamble_bytes);

    // Stub is now useless so delete it.
    // [csilvers: Commented out for perftools because it causes big problems
    //  when we're unpatching malloc.  We just let this live on as a leak.]
    //delete original_function_stub;

    // Restore the protection of the first MAX_PREAMBLE_STUB_SIZE bytes of
    // target to what they were before we started goofing around.
    succeeded = ::VirtualProtect(
        reinterpret_cast<void*>(target),
        MAX_PREAMBLE_STUB_SIZE,
        old_target_function_protect,
        &old_target_function_protect);

    // Flush the instruction cache to make sure the processor doesn't execute the
    // old version of the instructions (before our patch).
    //
    // See comment on FlushInstructionCache elsewhere in this file.
    succeeded = ::FlushInstructionCache(
        ::GetCurrentProcess(), target,  MAX_PREAMBLE_STUB_SIZE);
    if (!succeeded) {
        ASSERT(false, "Failed to flush instruction cache.");
        return SIDESTEP_UNEXPECTED;
    }

    return SIDESTEP_SUCCESS;
}


// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// Patch is a template for keeping the pointer to the original
// hooked routine, the function to call when hooked, and the
// stub routine which is patched.
template<class T>
class Patch 
{
public:
    // Constructor.  Does not hook the function yet.
    Patch<T>()
        : original_function_(NULL),
        patch_function_(NULL),
        stub_function_(NULL) 
    { }

    // Destructor.  Unhooks the function if it has been hooked.
    ~Patch<T>() 
    {
        Uninstall();
    }

    // Patches original function with func.
    // Must have called set_original to set the original function.
    void Install(T func) 
    {
        patch_function_ = func;

        assert(patch_function_ != NULL);
        assert(original_function_ != NULL);
        assert(stub_function_ == NULL);

        PreamblePatcher::Patch(original_function_,  patch_function_, &stub_function_);
    }

    // Un-patches the function.
    void Uninstall() 
    {
        if (stub_function_)
            PreamblePatcher::Unpatch(original_function_,
            patch_function_, stub_function_);
        stub_function_ = NULL;
    }

    // Set the function to be patched.
    void set_original(T original) { original_function_ = original; }

    // Get the original function being patched.
    T original() { return original_function_; }

    // Get the patched function.  (e.g. the replacement function)
    T patched() { return patch_function_; }

    // Access to the stub for calling the original function
    // while it is patched.
    T operator()() 
    {
        assert(stub_function_);
        return stub_function_;
    }

private:
    // The function that we plan to patch.
    T original_function_;
    // The function to replace the original with.
    T patch_function_;
    // To unpatch, we also need to keep around a "stub" that points to the
    // pre-patched Windows function.
    T stub_function_;
};

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// All Windows memory-allocation routines call through to one of these.
// DECLARE_PATCH(HeapCreate);
// DECLARE_PATCH(HeapDestroy);
DECLARE_PATCH(HeapAlloc);
DECLARE_PATCH(HeapReAlloc);
DECLARE_PATCH(HeapFree);


// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------
// Our replacement functions.
static LPVOID WINAPI Perftools_HeapAlloc(
    HANDLE hHeap, DWORD dwFlags, DWORD_PTR dwBytes) 
{
    LPVOID rv = patch_HeapAlloc()(hHeap, dwFlags, dwBytes);
    MemoryHook::m_sByteCnt += dwBytes;
    if (MemoryHook::m_sByteCnt > MemoryHook::m_sMaxByteCnt)
        MemoryHook::m_sMaxByteCnt = MemoryHook::m_sByteCnt;
    return rv;
}

// -----------------------------------------------------------------------------
static BOOL WINAPI Perftools_HeapFree(HANDLE hHeap, DWORD dwFlags, LPVOID lpMem)
{
    size_t size = 0;
    if (lpMem != 0) 
    {
        MemoryHook::m_sByteCnt -= HeapSize(hHeap, 0, lpMem);  // Will crash if lpMem is 0.
    }
    return patch_HeapFree()(hHeap, dwFlags, lpMem);
}

// -----------------------------------------------------------------------------
static LPVOID WINAPI Perftools_HeapReAlloc(
    HANDLE hHeap, DWORD dwFlags,  LPVOID lpMem, SIZE_T dwBytes) 
{
    // Don't call realloc, but instead do a free/malloc.  The problem is that
    // the builtin realloc may either expand a buffer, or it may simply
    // just call free/malloc.  If so, we will already have tracked the new
    // block via Perftools_HeapAlloc.

    LPVOID rv = Perftools_HeapAlloc(hHeap, dwFlags, dwBytes);

    // If there was an old buffer, now copy the data to the new buffer.
    if (lpMem != 0) 
    {
        size_t size = HeapSize(hHeap, 0, lpMem);
        if (size > dwBytes)
            size = dwBytes;
        memcpy(rv, lpMem, size);
        Perftools_HeapFree(hHeap, dwFlags, lpMem);
    }
    return rv;
}

// -----------------------------------------------------------------------------
bool MemoryHook::Hook() 
{
    if (!m_sIsHooked) 
    {
        // Luckily, Patch() doesn't call malloc or windows alloc routines
        // itself -- though it does call new (we can use PatchWithStub to
        // get around that, and will need to if we need to patch new).

        HMODULE hkernel32 = ::GetModuleHandle(L"kernel32");
        assert(hkernel32 != NULL);

        HMODULE hntdll = ::GetModuleHandle(L"ntdll");
        assert(hntdll != NULL);

        // Now that we've found all the functions, patch them

        // INSTALL_PATCH(HeapCreate);
        // INSTALL_PATCH(HeapDestroy);
        INSTALL_PATCH(HeapFree);
        INSTALL_PATCH(HeapReAlloc);
        INSTALL_PATCH(HeapAlloc);

        // We are finally completely hooked.
        m_sIsHooked = true;
    }
    return true;
}

// -----------------------------------------------------------------------------
bool MemoryHook::Unhook() 
{
    if (m_sIsHooked) 
    {
        // We need to go back to the system malloc/etc at global destruct time,
        // so objects that were constructed before tcmalloc, using the system
        // malloc, can destroy themselves using the system free.  This depends
        // on DLLs unloading in the reverse order in which they load!
        //
        // We also go back to the default HeapAlloc/etc, just for consistency.
        // Who knows, it may help avoid weird bugs in some situations.

        // UNINSTALL_PATCH(HeapCreate);
        // UNINSTALL_PATCH(HeapDestroy);
        UNINSTALL_PATCH(HeapAlloc);
        UNINSTALL_PATCH(HeapReAlloc);
        UNINSTALL_PATCH(HeapFree);

        m_sIsHooked = false;
    }
    return true;
}
#endif
