//
//  Thread safe for single Producer and single Consumer.
//  Uses no locks.
//
//  By Dennis Lang  Feb-2009
//
//  Simple fixed size ring buffer.
//  Manage objects by value.
//  Thread safe for single Producer and single Consumer (assuming the coping of 
//  the value in/out of the buffer is fast enough. When in doubt, store only a 
//  pointer to the value. Coping a pointer is probably an atomic operation and
//  should be thread safe.
//
//  Project: Thread optimized Ring Buffer 
//  Author: Dennis Lang
//  lang.dennis @ comcast.net 
//  http://home.comcast.net/~lang.dennis/

#pragma once

#include "IBuffer.h"

template <class T, size_t RingSize>
class RingBuffer : public IBuffer<T>
{
public:
    RingBuffer(size_t size = RingSize) 
        : m_size(size), m_buffer(new T[size]), m_rIndex(0), m_wIndex(0) 
        { assert(size > 1 && m_buffer != NULL); }

    ~RingBuffer() 
        { delete [] m_buffer; };

    size_t Next(size_t n) const 
        { return (n+1)%m_size; }
    bool Empty() const 
        { return (m_rIndex == m_wIndex); }
    bool Full() const
        { return (Next(m_wIndex) == m_rIndex); }

    bool Put(const T& value)
    {
        if (Full()) 
            return false;

        m_buffer[m_wIndex] = value;
        m_wIndex = Next(m_wIndex);

        return true;
    }

    bool Get(T& value)
    {
        if (Empty())
            return false;

        value = m_buffer[m_rIndex];
        m_rIndex = Next(m_rIndex);

        return true;
    }

private:
    T*              m_buffer;
    size_t          m_size;
    
    // volatile is only used to keep comipler from placing values in registers.
    // volatile does NOT make the index thread safe.
    char            filler1[16];    // spread data members to avoid thread locality penalty
    volatile size_t m_rIndex;
    char            filler2[16];    // spread data members to avoid thread locality penalty
    volatile size_t m_wIndex;
};
