//
//  Ring buffer (thread safe, multiple producers, consumers)
//  Stores data by value in ring buffer.
//
//  By Dennis Lang  Feb-2009
//
//  Simple fixed size ring buffer.
//  Manage objects by value.
//  Thread safe for multiple Producers and multiple Consumers.
//  Two locks, one for Producer and one for Consumer.
//
//  Project: Thread optimized Ring Buffer 
//  Author: Dennis Lang
//  lang.dennis @ comcast.net 
//  http://home.comcast.net/~lang.dennis/

#pragma once
#include "Lock.h"
#include "IBuffer.h"

template <class T, size_t RingSize, class L>
class RingBufferLock : public IBuffer<T>
{
public:
    RingBufferLock(size_t size = RingSize) 
        : m_buffer(new T[size]), m_size(size), m_rIndex(0), m_wIndex(0) 
        { assert(size > 1 && m_buffer != NULL); }

    ~RingBufferLock()
        { delete [] m_buffer; };

    size_t Next(size_t n) const 
        { return (n+1)%m_size; }
    bool Empty() const 
        { return (m_rIndex == m_wIndex); }
    bool Full() const
        { return (Next(m_wIndex) == m_rIndex); }

    bool Put(const T& value)
    {
        LockIt lockit(m_wLock);
        if (Full()) 
            return false;

        m_buffer[m_wIndex] = value;
        m_wIndex = Next(m_wIndex);
        return true;
    }

    bool Get(T& value)
    {
        LockIt lockit(m_rLock);
        if (Empty())
            return false;
        value = m_buffer[m_rIndex];
        m_rIndex = Next(m_rIndex);
        return true;
    }

private:
    T*              m_buffer;
    size_t          m_size;

    // volatile is only used to keep comipler from placing values in registers.
    // volatile does NOT make the index thread safe.

    volatile size_t m_rIndex;
    L m_rLock;

    volatile size_t m_wIndex;
    L m_wLock;
};
