//  StackMS.h
//  Thread safe Stack template using  MSDN's Singly Linked List.
//
//  Project: Thread optimized Stack Buffer 
//  Author: Dennis Lang
//  lang.dennis @ comcast.net 
//  http://home.comcast.net/~lang.dennis/

#pragma once
#include "Lock.h"
#include "IBuffer.h"

// Fixed sized thread save Stack using Mircosoft's Sinlgy Linked List API.
//
// Create two Linked List (free and data).
// Free is filled at construction and used to provide containers to hold data added by caller to data list.
// When data is popped off the container is placed back on the free list.
//
// T = data type
// RingSize = ring size
template <class T, size_t RingSize>
class StackMS : public IBuffer<T>
{
private:
// we need to align the memory-pool-chunks.
#pragma pack(push, MEMORY_ALLOCATION_ALIGNMENT)

    /// List Item
    template <typename TYPE>
    struct ListItem
    {
        /// Next entry in the singly-linked list.
        SLIST_ENTRY Next;
        T value;
    };
    typedef ListItem<T> ListValue;

#pragma pack(pop, MEMORY_ALLOCATION_ALIGNMENT)

    /// Head of free linked list.
    SLIST_HEADER m_freeList;
    ListValue m_freeValues[RingSize];

    /// Head of data linked list.
    SLIST_HEADER m_dataList;

    // no copying is supported
    StackMS& operator=(const StackMS&);
    StackMS(const StackMS&);

public:
    StackMS()
    {
        InitializeSListHead(&m_freeList);
        for (size_t i = 0; i < RingSize; i++)
        {
            InterlockedPushEntrySList(&m_freeList, &m_freeValues[i].Next);
        }
        InitializeSListHead(&m_dataList);
    }

    ~StackMS()
    {
        InterlockedFlushSList(&m_freeList);
        InterlockedFlushSList(&m_dataList);
    }
    
    size_t Size() 
    {
        return QueryDepthSList(&m_dataList);
    }

    // Get object off of stack (LIFO)
    bool Get(T& value)
    {
        ListValue* pListValue = reinterpret_cast<ListValue*>(InterlockedPopEntrySList(&m_dataList));
        if (NULL == pListValue)
            return false;
        value = pListValue->value;
        InterlockedPushEntrySList(&m_freeList, &pListValue->Next);
        return true;
    }

    // Put object on of stack (LIFO)
    bool Put(const T& value)  
    {
        ListValue* pListValue = reinterpret_cast<ListValue*>(InterlockedPopEntrySList(&m_freeList));
        if (NULL == pListValue)
            return false;

        pListValue->value = value;
        InterlockedPushEntrySList(&m_dataList, &pListValue->Next);
        return true;
    }

    // Reverse order of Linked List items.
    ListValue* ReverseSList(ListValue* pHead)
    {
	    ListValue * pPrev = NULL;
        ListValue * pThis = pHead;
	    ListValue * pNext;  

   	    while (pThis)
        {
		    pNext = pThis->Next;
		    pThis->Next = pPrev;
		    pPrev = pThis;
		    pThis = pNext;
	    }

	    return pPrev;
    }

    // Get all items in LIFO order.
    ListValue* GetAll()
    {
        ListValue* pItems = InterlockedFlushSList(&m_dataList);
        PutAllOnFree(pItems);
        return pItems;
    }

    // Get all items in FIFO order.
    ListValue* GetAllReverse()
    {
        ListValue* pItems = ReverseSList(InterlockedFlushSList(&m_dataList));
        PutAllOnFree(pItems);
        return pItems;
    }

    void PutAllOnFree(ListValue* pItems)
    {
        while (pItems != NULL)
        {
            ListItem* pNextItem = pItems->Next;
            InterlockedPushEntrySList(&m_freeList, pItems);
            pItems = pNextItem;
        }
    }
};

 