//
// Author: Dennis Lang
// lang.dennis @ comcast.net 
// http://home.comcast.net/~lang.dennis/
//
//
// Jan 2015 - Ring buffer performance.
// ================================== 
// 
// Requirements:
//    1. A data structure which can be shared by two or more threads. 
//
// The following code is a collect of Queue classes from a DDJ artile by Herb Sutter and a 
// few Ring buffers I wrote. The 'ring_spin_test.cpp' is the main application which
// will run one or more of the container classes through several iterations to measure their 
// 'transaction' performance.  The test iteration varies the 'Data' object size and the number
// of Producer and Consumer threads.  
// 
// Container objects:
//    RingBuffer              LockFree ring buffer, only works with 1 producer and 1 consumer thread.
// Following are thread safe:
//    RingBufferLock/Pad      Ring buffers store objects by value, one with Locality padding.
//    RingBufferPtrLock/Pad   Ring buffers which allocate storage on the heap and store, one with padding.
//    StackMS                 Stack using Mircosoft's Sinlgy Linked List API.
//    boost::lockfree::queue  Boost Lock-free queue
// 
// 
// The variations ending in pad include padding between class data memebers 
// to avoid thread locality penalty.  See http://home.comcast.net/~lang.dennis/code/locality.html
// 
// An Okay or Error status is reported at the end of each test collection to report whether or
// not the Producer and the Consumer processed the exact same data.  The verification is just the sumation
// of the data object values.
// 
// Based on my results, the Ring buffers can out perform a queue and has the added advantage of allowing
// the designer to limit the maximum resources used.
// 
// 
// Dr Dobbs articles on thread safe queues and queue performance:
// 
// # "Lock-Free Queues", by P. Marginean, DDJ July 2008
// # Writing Lock-Free code: A Corrected Queue, by Herb Sutter, DDJ September 2008
// # Maximize Locality, Minimize Contention, by Herb Sutter, May 2008
// # "Lock-Free Code: A False Sense of Security", by Herb Sutter, DDJ October 2008
// # "Measuring Parallel Performance: Optimizing a Concurrent Queue", by Herb Sutter, DDJ Jan 2009 
//
//

#include "stdafx.h"
#include <windows.h>
#include <assert.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <functional>
#include <numeric>
#include <math.h>

#include "comma.h"
#include "Lock.h"

#include "RingBuffer.h"

#include "RingBufferLock.h"
#include "RingBufferLockPad.h"

#include "RingBufferPtrLock.h"
#include "RingBufferPtrLockPad.h"

#include "StackMS.h"

// Boost
#define TEST_BOOST
#ifdef TEST_BOOST
#define BOOST_SYSTEM_NO_DEPRECATED

#include <boost/version.hpp>
#include <boost/lockfree/queue.hpp>

// Boost lockfree queue is thread safe. 
template <typename TT, size_t RingSize>
class BoostFixedSzQueue : public boost::lockfree::queue<TT, boost::lockfree::fixed_sized<true>>
{
public:
    BoostFixedSzQueue(size_t size = RingSize)
        : queue(RingSize)
    {     assert(size > 1); }
    bool Empty() const
    {  return empty();  }
    // bool Full() const
    // {  return full();  }

    bool Put(const TT& value)
    {  return bounded_push(value);  }

    bool Get(TT& value)
    { return pop(value); }
};
#endif

#include "HeapHook.h"

bool g_verbose = false;
volatile bool g_bRunProducer = true;
volatile bool g_bRunConsumer = true;
const size_t sBusyBufSize = 1024*3;

// -----------------------------------------------------------------------------
struct ThreadStats
{
    ThreadStats() : counter(0), sum(0), feCnt(0) {}
    size_t counter;
    size_t sum;
    size_t feCnt;
};

// -----------------------------------------------------------------------------
template <class BufType>
struct ThreadInfo
{
    ThreadInfo() : id(0), hnd(INVALID_HANDLE_VALUE), stats() {}

    DWORD       id;
    HANDLE      hnd;
    ThreadStats stats;
    BufType*    pBuffer;    // ring, queue buffer object
    char        pad[512];
};

// -----------------------------------------------------------------------------
// Producer test thread
template <class T, class V>
DWORD Producer(ThreadInfo<T>& tInfo)
{
    V value = 0;
    ThreadStats stats;
    char  keepBusyPtr[sBusyBufSize];

    while (g_bRunProducer)
    {
        // Generate a value 
        value = (value+1) & 0x7; // value = rand();

        // Do some local work to touch  memory.
        if (sBusyBufSize != 0)
            memset(keepBusyPtr, (int)value, sBusyBufSize);

        if (!tInfo.pBuffer->Put(value))
        {
            stats.feCnt++;  // full count
            continue;
        }

        stats.sum += value;
        stats.counter++;
    }


    // Save results.
    tInfo.stats = stats;
    return tInfo.id;
}

// -----------------------------------------------------------------------------
// Consumer test thread
template <class T, class V>
DWORD Consumer(ThreadInfo<T>& tInfo)
{
    V value = 0;
    ThreadStats stats;
    char keepBusyPtr[sBusyBufSize];

    while (g_bRunConsumer)
    {
        if (tInfo.pBuffer->Get(value))
        {
            stats.sum += value;
            stats.counter++;
        }
        else
        {
            stats.feCnt++;
        }

        // Do some local work which locals cpu cache.
        if (sBusyBufSize != 0)
            memset(keepBusyPtr, (int)value, sBusyBufSize);
    }

    // Drain any residual values.
    // Sleep(10);     // Wait to give producer chance to finished.
    while (tInfo.pBuffer->Get(value))
    {
        stats.sum += value;
        stats.counter++;
    }

    // Save results.
    tInfo.stats = stats;
    return tInfo.id;
}


// -----------------------------------------------------------------------------
// Copy input upto endAt into destination buffer
const char* LeftTo(char* dstBuf, int maxBuf, const char* inStr, const char* endAt)
{
    size_t spaceIdx = strcspn(inStr, " ");
    if (spaceIdx < 10)
        inStr = inStr + spaceIdx;
    const char* endPtr = strstr(inStr, endAt);
    if (endPtr)
    {
        size_t cpySize = endPtr - inStr;
        strncpy_s(dstBuf, maxBuf, inStr, cpySize);
        return dstBuf;
    }

    return inStr;
}

// -----------------------------------------------------------------------------
__int64 GetClockTics()
{
    LARGE_INTEGER counter;
    QueryPerformanceCounter(&counter);
    return counter.QuadPart;
}

__int64 TicsPerSecond()
{
    LARGE_INTEGER freq;
    QueryPerformanceFrequency(&freq);
    return freq.QuadPart;
}

// -----------------------------------------------------------------------------
typedef std::vector<double>  DblVecList;

template <typename Tvalue>
Tvalue Average(std::vector<Tvalue>& tlist)
{
    std::sort(tlist.begin(), tlist.end());

    // Skip smallest and largest.
    double sum = 0;
    for (unsigned idx = 1; idx != tlist.size()-1; idx++)
        sum += tlist[idx];

    return sum / (tlist.size()-2);
}

template <typename Titer>
double StdDevPer(Titer begin, Titer end, size_t length)
{
    // Compute relative standard deviation percent
    double mean = accumulate(begin, end, 0.0) / length;
    std::vector<double> zero_mean(begin, end);
    std::transform(zero_mean.begin(), zero_mean.end(), zero_mean.begin(), bind2nd(std::minus<double>(), mean ) );

    double deviation = std::inner_product(zero_mean.begin(), zero_mean.end(), zero_mean.begin(), 0.0 );
    deviation = sqrt( deviation / (length-1) );
    return deviation / mean;
}

template <typename TT>
TT BitOnCnt(TT value)
{
    TT cnt = 0;
    for (TT bit = 1; bit < value; bit *= 2)
    {
        if ((value & bit) == bit)
            cnt++;
    }
    return cnt;
}

template <typename TT>
double BestStdDevPer(TT& list, size_t minSetSize)
{
    unsigned listSz = list.size();
    double minStd = 100;
    unsigned minBits = 0;

    unsigned bits = 0;
    bool more = true;
    while (more)
    {
        bits++;
        unsigned bitOnCnt = 0;
        while ((bitOnCnt = BitOnCnt(bits)) < minSetSize)
            bits++;
        more = bitOnCnt != listSz;

        TT experiment;
        unsigned idx = 0;
        for (unsigned b = 1; b <= bits; b *= 2, idx++)
        {
            if ((bits & b) == b)
                experiment.push_back(list[idx]);
        }

        double stdDevPer = StdDevPer(experiment.begin(), experiment.end(), experiment.size());
        if (stdDevPer < minStd)
        {
            minStd = stdDevPer;
            minBits = bits;
        }
    }

    TT experiment;
    unsigned idx = 0;
    for (unsigned b = 1; b <= minBits; b *= 2, idx++)
    {
        if ((minBits & b) == b)
            experiment.push_back(list[idx]);
    }

    list.swap(experiment);
    return minStd;
}

// -----------------------------------------------------------------------------
template <class BufType, class ValType>
void Test(size_t nProducer=1, size_t nConsumer=1, size_t seconds=5)
{
   

    //
    //  Test Buffer, by running multiple threads:
    //  Producer thread(s), which fills the buffer and
    //  Consumer thread(s) which drains the buffer.
    //


    size_t nThreads = nProducer + nConsumer; 
  
    DblVecList transPerSecList;
    size_t memLeakCnt = 0;
    size_t memMaxCnt = 0;
    bool badBuffer = false;
    const unsigned sNumSamples = 5;
    const unsigned sMinStdSetSize = 3;
    const double sMinStdDevPer = 2.0;
    double stdDevPer;

    do
    {
        memLeakCnt = 0;
        memMaxCnt = 0;
        badBuffer = false;
        transPerSecList.clear();

        while (transPerSecList.size() != sNumSamples)
        {
            // Clear stats
            ThreadInfo<BufType>* pTinfo = new ThreadInfo<BufType>[nThreads];

            MemoryHook::ResetByteCnt();
            // Create buffer shared by threads.
            BufType buffer;
            memMaxCnt = MemoryHook::GetMaxByteCnt();

            // Create producer threads
            size_t tIdx = 0;
            for (size_t pIdx = 0; pIdx < nProducer; pIdx++)
            {
                pTinfo[tIdx].pBuffer = &buffer;
                pTinfo[tIdx].hnd =
                    CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) Producer<BufType, ValType>, &pTinfo[tIdx], CREATE_SUSPENDED, &pTinfo[tIdx].id);
                SetThreadPriority(pTinfo[tIdx].hnd, THREAD_PRIORITY_HIGHEST);
                tIdx++;
            }

            // Create consumer threads
            for (size_t cIdx = 0; cIdx < nConsumer; cIdx++)
            {
                pTinfo[tIdx].pBuffer = &buffer;
                pTinfo[tIdx].hnd =
                    CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) Consumer<BufType, ValType>, &pTinfo[tIdx], CREATE_SUSPENDED, &pTinfo[tIdx].id);
                SetThreadPriority(pTinfo[tIdx].hnd, THREAD_PRIORITY_HIGHEST);
                tIdx++;
            }

            MemoryHook::ResetByteCnt(0, memMaxCnt);

            // Start threads
            __int64 startTics = GetClockTics();
            g_bRunProducer = g_bRunConsumer = true;
            for (tIdx = 0; tIdx < nThreads; tIdx++)
            {
                ResumeThread(pTinfo[tIdx].hnd);
            }

            // Run test for 'seconds'.
            Sleep(DWORD(seconds * 1000));
            g_bRunProducer = false;

            for (tIdx = 0; tIdx < nProducer; tIdx++)
            {
                DWORD id = WaitForSingleObject(pTinfo[tIdx].hnd, INFINITE);
            }

            Sleep(0);
            g_bRunConsumer = false;
            for (tIdx = nConsumer; tIdx < nThreads; tIdx++)
            {
                DWORD id = WaitForSingleObject(pTinfo[tIdx].hnd, INFINITE);
            }

            // Report results
            __int64 delta = GetClockTics() - startTics;
            double elapsedSeconds = (delta * 1000 / TicsPerSecond()) * 1e-3;

            memLeakCnt = MemoryHook::GetByteCnt();
            memMaxCnt = MemoryHook::GetMaxByteCnt();

            for (tIdx = nConsumer; tIdx < nThreads; tIdx++)
            {
                CloseHandle(pTinfo[tIdx].hnd);
            }

            if (g_verbose)
            {
                std::cout << typeid(buffer).name() << std::endl;
                std::cout << "Tics=" << delta << " sec=" << elapsedSeconds << std::endl;
            }

            size_t prodTotCount = 0;
            LONGLONG prodTotSum = 0;
            tIdx = 0;
            for (DWORD pIdx = 0; pIdx < nProducer; pIdx++, tIdx++)
            {
                if (g_verbose)
                    std::cout << "  Producer count/sec=" << std::fixed << std::setprecision(0)
                    << pTinfo[tIdx].stats.counter / elapsedSeconds
                    << " data sum=" << pTinfo[tIdx].stats.sum
                    << " fullCnt= " << pTinfo[tIdx].stats.feCnt
                    << std::endl;
                prodTotCount += pTinfo[tIdx].stats.counter;
                prodTotSum += pTinfo[tIdx].stats.sum;
            }
            if (nProducer > 1 && g_verbose)
                std::cout << "T Producer count/sec=" << std::fixed << std::setprecision(0)
                << prodTotCount / elapsedSeconds
                << " data sum=" << prodTotSum
                << std::endl
                << std::endl;

            size_t consTotCount = 0;
            LONGLONG consTotSum = 0;
            for (DWORD cIdx = 0; cIdx < nConsumer; cIdx++, tIdx++)
            {
                if (g_verbose)
                    std::cout << "  Consumer count/sec=" << std::fixed << std::setprecision(0)
                    << pTinfo[tIdx].stats.counter / elapsedSeconds
                    << " data sum=" << pTinfo[tIdx].stats.sum
                    << " emptyCnt=" << pTinfo[tIdx].stats.feCnt
                    << std::endl;
                consTotCount += pTinfo[tIdx].stats.counter;
                consTotSum += pTinfo[tIdx].stats.sum;
            }
            if (nConsumer > 1 && g_verbose)
                std::cout << "T Consumer count/sec=" << std::fixed << std::setprecision(0)
                << consTotCount / elapsedSeconds
                << " data sum=" << consTotSum
                << std::endl
                << std::endl;

            if (prodTotSum == consTotSum)
                transPerSecList.push_back(prodTotCount / elapsedSeconds);
            else
            {
                badBuffer = true;
                transPerSecList.push_back(0);
            }

            delete [] pTinfo;
        }
    
        stdDevPer = BestStdDevPer(transPerSecList, sMinStdSetSize) * 100;
        // stdDevPer = StdDevPer(transPerSecList.begin(), transPerSecList.end(), transPerSecList.size()) * 100;
        if (stdDevPer >= sMinStdSetSize)
            std::cerr << stdDevPer << "\n";
    } while (!badBuffer && stdDevPer > sMinStdDevPer);    // Loop until stable results

    double transPerSec = Average(transPerSecList);

    char nameBuf[256];
    if (!g_verbose)
        std::cout 
  //          << "  " << std::setw(5) << elapsedSeconds
            << std::setw(10) << std::fixed << std::setprecision(0) << transPerSec
            << ", " << std::setw(5) << nProducer
            << ", " << std::setw(5) << nConsumer
            << ", " << std::setw(8) << sizeof(ValType)
            << (badBuffer ? ",  ERROR" : ",  okay ")
			<< ", " << std::setw(7) << memLeakCnt
			<< ", " << std::setw(7) << memMaxCnt
            << ", " << std::left << std::setw(22) << LeftTo(nameBuf, sizeof(nameBuf), typeid(BufType).name(), "<") << std::right;

    if (transPerSecList.size() > 1)
    {
        std::cout << ", " << std::setprecision(0) << std::setw(3) << StdDevPer(transPerSecList.begin(), transPerSecList.end(), transPerSecList.size()) * 100;
        for (size_t idx = 0; idx != transPerSecList.size(); idx++)
            std::cout << ", " << transPerSecList[idx];

    }

    std::cout << std::endl;
}

// -----------------------------------------------------------------------------
// Data object with variable payload size.
template <int nSize>
struct DataObj
{
public:
    DataObj() {}
  
    DataObj(size_t rhs)
        { memset(str, (int)rhs, sizeof(str));  n = rhs;}

    DataObj& operator=(size_t rhs)
        { memset(str, (int)rhs, sizeof(str));  n = rhs; return *this;}

    DataObj& operator++()
        { n++; return *this;}

    operator size_t() 
        { return n; }

private:
    size_t n;
    char str[nSize];
};

 
// -----------------------------------------------------------------------------
#ifdef _M_X64
#include  <intrin.h>
#endif

float GetProcessorMhz()
{
#ifdef _M_X64
    LARGE_INTEGER    s, e, freq;
    unsigned __int64 cs, ce;

    // Determine timer frequency.
    QueryPerformanceFrequency(&freq);

    QueryPerformanceCounter(&s);
    cs = __rdtsc();

    // Wait for a while...
    for (volatile long i = 0; i < 1000000; ++i) ;

    ce = __rdtsc(); 
    QueryPerformanceCounter(&e);

    // Calculate frequency.
    double cpu_speed = ((ce - cs) * (double)freq.QuadPart / (e.QuadPart - s.QuadPart));
    return (float)cpu_speed;
#else
    /*
    http://www.codeproject.com/KB/system/Processor_Speed.aspx?fid=59743&df=90&mpp=25&noise=3&sort=Position&view=Quick&select=849085

    RdTSC:
    It's the Pentium instruction "ReaD Time Stamp Counter". It measures the
    number of clock cycles that have passed since the processor was reset, as a
    64-bit number. That's what the <CODE>_emit lines do.*/
    #define RdTSC __asm _emit 0x0f __asm _emit 0x31

    // variables for the clock-cycles:
    __int64 cyclesStart = 0, cyclesStop = 0;
    // variables for the High-Res Preformance Counter:
    LARGE_INTEGER nFreq, nCtrStop, nCtr;

    // retrieve performance-counter frequency per second:
    if(!QueryPerformanceFrequency(&nFreq)) 
        return 0;

    // retrieve the current value of the performance counter:
    QueryPerformanceCounter(&nCtrStop);

    // add the frequency to the counter-value:
    nCtrStop.QuadPart += nFreq.QuadPart;


    _asm
        {// retrieve the clock-cycles for the start value:
            RdTSC
            mov DWORD PTR cyclesStart, eax
            mov DWORD PTR [cyclesStart + 4], edx
        }

        do{
        // retrieve the value of the performance counter
        // until 1 sec has gone by:
             QueryPerformanceCounter(&nCtr);
          }while (nCtr.QuadPart < nCtrStop.QuadPart);

    _asm
        {// retrieve again the clock-cycles after 1 sec. has gone by:
            RdTSC
            mov DWORD PTR cyclesStop, eax
            mov DWORD PTR [cyclesStop + 4], edx
        }

    // stop-start is speed in Hz divided by 1,000,000 is speed in MHz
    return ((float)cyclesStop-(float)cyclesStart) / 1000000;
#endif

}


// -----------------------------------------------------------------------------
float GetRegProcessorMhz()
{
    DWORD dwMHz = _MAX_PATH;
    DWORD BufSize = sizeof(dwMHz);

    // open the key where the proc speed is hidden:
    HKEY hKey;
    long lError = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
            L"HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
            0,
            KEY_READ,
            &hKey);

    // query the key:
    RegQueryValueEx(hKey, L"~MHz", NULL, NULL, (LPBYTE) &dwMHz, &BufSize);
    return (float)dwMHz;
}

// -----------------------------------------------------------------------------
template 
    <unsigned int DATA_SIZE,    // payload size
    unsigned int RING_SIZE,     // ring size (max entries before blocking producer)
    unsigned int nMaxThreads,   // max consumer or producer threads
    unsigned int nSeconds>      // duration of test in seconds
void TestBuffers()
{
    std::cout << "DataSize=" << DATA_SIZE << "  RingSize=" << RING_SIZE << "  Seconds=" << nSeconds << std::endl;
    Test<RingBuffer<DataObj<DATA_SIZE>, RING_SIZE>, DataObj<DATA_SIZE>>(1, 1, nSeconds);
    
    for (unsigned int nProducer = 1; nProducer <= nMaxThreads; nProducer++)
    {
        for (unsigned int nConsumer = 1; nConsumer <= nMaxThreads; nConsumer++)
        {
            Test<StackMS<DataObj<DATA_SIZE>, RING_SIZE>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);

#ifdef TEST_BOOST
            Test<BoostFixedSzQueue<DataObj<DATA_SIZE>, RING_SIZE>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
#endif

            Test<RingBufferLock<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
            Test<RingBufferLockPad<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);

            Test<RingBufferPtrLock<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
            Test<RingBufferPtrLockPad<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
        }
        std::cout << "\n";
        std::cerr << nProducer*100/nMaxThreads << "%\n";
    }
}

// -----------------------------------------------------------------------------
template
<unsigned int DATA_SIZE,    // payload size
unsigned int RING_SIZE,     // ring size (max entries before blocking producer)
unsigned int nMaxThreads,   // max consumer or producer threads
unsigned int nSeconds>      // duration of test in seconds
void TestLocalityPenalty()
{
    int nProducer = 1;
    int nConsumer = 1;
    Test<RingBufferLock<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
    Test<RingBufferLockPad<DataObj<DATA_SIZE>, RING_SIZE, Interlock>, DataObj<DATA_SIZE>>(nProducer, nConsumer, nSeconds);
 
    std::cout << "\n";
}

// -----------------------------------------------------------------------------
int main(int argc, char* argv [])
{
    const size_t nMaxThreads = 2;
    const size_t nSeconds = 4;
    const size_t RING_SIZE = 100;

    g_verbose = (argc > 1);

    char computerName[MAX_PATH];
    DWORD cnSize = sizeof(computerName);
    GetComputerNameA(computerName, &cnSize);
    SYSTEM_INFO systemInfo;
    GetSystemInfo(&systemInfo);

    std::ostringstream ostr;
    ostr << "System=" << computerName
        << ", #Processors=" << systemInfo.dwNumberOfProcessors
        << ", ProcGhz=" << GetRegProcessorMhz() / 1000.0
        << ", ProcType=" << systemInfo.dwProcessorType
        << ", ProcArch=" << systemInfo.wProcessorArchitecture
        << ", ProcLevel=" << systemInfo.wProcessorLevel;

    ostr << ", RingSize=" << RING_SIZE
        << ", MaxThreads=" << nMaxThreads
        << ", Seconds=" << nSeconds
        << "\n";

    // std::cerr << ostr.str();
    std::cout << ostr.str();

    std::cerr << "Hit return to start test (recommend setting priority high on this exe) ";
    char buf[2];
    fgets(buf, sizeof(buf), stdin);

    if (!g_verbose)
        std::cout << " Trans/sec, #Prod, #Cons, DataSize,   Okay, MemLeak,  MaxMem,  TestName, StdDev%, Trans/sec samples...\n";
    else
        EnableCommaCout();

    if (false)
    {
        // 
        //  Run buffer performance tests with various size payloads.
        //
        const unsigned int repeatCnt = 1;
        const unsigned int nnSeconds = 3;
        std::cout << "\nTest Locality Penalty for " << nnSeconds << " seconds\n";
        for (unsigned int repeatIdx = 0; repeatIdx < repeatCnt; repeatIdx++)
        {
            TestLocalityPenalty<1, RING_SIZE, nMaxThreads, nnSeconds>();
            TestLocalityPenalty<20, RING_SIZE, nMaxThreads, nnSeconds>();
            TestLocalityPenalty<200, RING_SIZE, nMaxThreads, nnSeconds>();
        }
        std::cout << std::endl;
    }

    if (true)
    {
        // 
        //  Run buffer performance tests with various size payloads.
        //
        const unsigned int repeatCnt = 1;
        std::cout << "\nTest Performance of varing thread count for " << nSeconds << " seconds\n";
        for (unsigned int repeatIdx = 0; repeatIdx < repeatCnt; repeatIdx++)
        {
            TestBuffers<1, RING_SIZE, nMaxThreads, nSeconds>();
            TestBuffers<20, RING_SIZE, nMaxThreads, nSeconds>();
            TestBuffers<200, RING_SIZE, nMaxThreads, nSeconds>();
        }
        std::cout << std::endl;
    }

    if (false)
    {
        std::cout << "\nTest Performance of varing ring size for " << nSeconds << " seconds\n";
        TestBuffers<1, RING_SIZE, nMaxThreads, nSeconds>();
        TestBuffers<1, RING_SIZE * 10, nMaxThreads, nSeconds>(); 
        TestBuffers<1, RING_SIZE*100, nMaxThreads, nSeconds>();
        std::cout << std::endl;
    }
}
